/*

   Adapted from code by Yuri Volobuev (volobuev@t1.chem.umn.edu)
                    and Mike Muuss (author of ping, email address unknown)

   Send a broadcast ICMP_ECHO packet with a bogus ethernet MAC address, the
   idea being to detect hosts which are running their interfaces in
   promiscuous mode (packet sniffers, Apple Macs, etc.)

   Chris Lightfoot <cwrl2@hermes.cam.ac.uk>    1998

   The code is in the public domain.

*/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <netdb.h>
#include <sys/socket.h>

#include <arpa/inet.h>

#include <net/if.h>

#include <signal.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <linux/sockios.h>
#include <linux/if_ether.h>

#define DEFAULT_DEVICE	"eth0"
#define PROGNAME	"findpromisc"

#define MACADDR_LEN	6
#define IPADDR_LEN	4
#define DATA_LEN	56
#define IPVERSION	4

#define RECV_BUFFER_LEN	1500

// perform an IP checksum
unsigned short ip_checksum(char* a,char len)
{
	long sum = 0;
	unsigned short *addr = (unsigned short*)a;

	while(len > 1)
	{
		sum += *addr++;
		len -= 2;
	}

	if(len > 0) sum += *addr;
	while (sum>>16) sum = (sum & 0xffff) + (sum >> 16);

	return ~sum;
}

// return an IP address in network order
int get_ip_addr(char *address, unsigned long *na)
{
	struct hostent *host;
	static struct in_addr saddr;

	// First try it as aaa.bbb.ccc.ddd
	if (inet_aton(address, &saddr))
	{
		*na = saddr.s_addr;
		return 1;
	}

	host = gethostbyname(address);

	if (host != NULL)
	{
		*na = ((struct in_addr*)*host->h_addr_list)->s_addr;
		return 1;
	}
		
	return 0;
}

// return a dotted IP address from an unsigned long in network order
char ip_string[32];

char *ipaddr_to_string(unsigned long a)
{
/*	sprintf(ip_string, "%d.%d.%d.%d",
			(unsigned long)(a >>  0) & 0xff,
			(unsigned long)(a >>  8) & 0xff,
			(unsigned long)(a >> 16) & 0xff,
			(unsigned long)(a >> 24) & 0xff);

	return ip_string;*/

	struct in_addr in;

	in.s_addr = a;

	return inet_ntoa(in);
}

char hw_string[6 * 2 + 7];

char *hwaddr_to_string(char *hwaddr)
{
	sprintf(hw_string, "%02x:%02x:%02x:%02x:%02x:%02x",
			(unsigned long)hwaddr[0] & 0xff,
			(unsigned long)hwaddr[1] & 0xff,
			(unsigned long)hwaddr[2] & 0xff,
			(unsigned long)hwaddr[3] & 0xff,
			(unsigned long)hwaddr[4] & 0xff,
			(unsigned long)hwaddr[5] & 0xff);

	return hw_string;
}

char string[] = "Hello Mr. Flibble!";

struct sockaddr sa;
int sock, packetlen;
char *packetdata, *recvdata;
int npings;
int verbose = 0, broadcast = 0, help = 0;

// deals with actually sending stuff, signal handler
void pinger(int n)
{
	if (n != SIGALRM)
		return; // can't happen

	if (sendto(sock, packetdata, packetlen, 0, &sa, sizeof(sa)) < 0)
	{
		perror (PROGNAME": sendto");
		exit(1);
	}

/*	if (send(sock, packetdata, packetlen, 0) < 0)
	{
		perror(PROGNAME": send");
		exit(1);
	}*/

	if (verbose) printf(">>> sent packet\n");

	npings++;

	alarm(1);
}

int 
main (int argc, char **argv)
{
	int i, n, m;
	int icmpsock;
	struct protoent *proto;
	struct sockaddr_in sa_from;
	int buffersize = 65536;
	struct icmp *icmpreply;

	char **p;

	struct in_addr si;
	unsigned long destaddr = 0;
	int gotaddress = 0;
	unsigned long sourceaddr;
	char source_macaddr[MACADDR_LEN];

	char if_name[IFNAMSIZ];

	struct ethhdr	*eth;
	struct iphdr	*ip;
	struct icmp	*icmp;
	char *data;

	struct ifreq ifr;

	// <--- start processing arguments
	strcpy(if_name, DEFAULT_DEVICE);

	for (p = (argv + 1); p < (argv + argc); p++)
	{
		if (!strcmp(*p, "-h")) help = 1;

		else if (!strcmp(*p, "-v")) verbose = 1;

		else if (!strcmp(*p, "-i"))
		{
			if ((p + 1) >= (argv + argc))
			{
				fprintf(stderr,
					PROGNAME": you must specify an interface after -i\n");
				return 1;
			}

			else
				strncpy(if_name, *(++p), IFNAMSIZ);
		}

		else if (!strcmp(*p, "-b")) broadcast = 1;

		else if (!strcmp(*p, "-l"))
		{
			destaddr = 0;
			gotaddress = 1;
		}
		
		else
		{
			if (gotaddress)
			{
				fprintf(stderr, PROGNAME": you can specify only one of -b, -l and a target address\n");
			}
			if (!get_ip_addr(*p, &destaddr))
			{
				fprintf(stderr, PROGNAME": unknown or invalid host `%s'\n", *p);
				return 1;
			}
			else
				gotaddress = 1;
		}
	}

	if (argc < 2 || help)
	{
		fprintf(stderr, PROGNAME": detect promiscuous-mode interfaces on other hosts\n"
				"\n"
				"\t"PROGNAME" [-h] [-v] [-i interface] {-b, -l, address}\n"
				"\n"
				"\t-h\t\tdisplay this help message\n"
				"\t-v\t\tbe verbose\n"
				"\t-i interface\tuse interface specified (default `%s')\n"
				"\t-b\t\tuse the broadcast address\n"
				"\t-l\t\tuse `local broadcast' address 0.0.0.0\n"
				"\taddress\t\taddress to ping\n\n", DEFAULT_DEVICE);

		return 0;
	}

	if (!gotaddress && !broadcast)
	{
		fprintf(stderr, PROGNAME": you must either specify an address or -b or -l for broadcast\n");
		return 1;
	}

	// finish processing arguments --->

	// <--- set up the sockets required

	// get a socket capable of sending raw ethernet frames
	sock = socket(AF_INET, SOCK_PACKET, htons(ETH_P_RARP));

	if (sock < 0)
	{
		perror(PROGNAME": socket");
		return 1;
	}

	// record the interface address
	memset(&sa, 0, sizeof(sa));
	strcpy(sa.sa_data, if_name);

	if (verbose)
	{
		fprintf(stderr, PROGNAME": using interface `%s'\n", if_name);
	}

	// get a socket for receiving icmp responses
	proto = getprotobyname("icmp");

	if (proto == NULL)
	{
		fprintf(stderr, PROGNAME": icmp: unknown protocol\n");
		return 2;
	}

	icmpsock = socket(AF_INET, SOCK_RAW, proto->p_proto);

	if (icmpsock < 0)
	{
		perror(PROGNAME": socket");
		return 1;
	}

	recvdata = (char*)malloc(RECV_BUFFER_LEN);
	
	if (!recvdata)
	{
		perror(PROGNAME": malloc");
		return 1;
	}

	// finish setting up sockets --->

	// <--- set up the ICMP_ECHO packet to be sent

	// allocate space for the packet to send
	packetdata = (char*)malloc(packetlen = (sizeof(struct ethhdr) +
				   sizeof(struct iphdr) +
				   sizeof(struct icmp) +
				   DATA_LEN));

	if (!packetdata)
	{
		perror(PROGNAME": malloc");
		return 1;
	}

	if (verbose)
	{
		fprintf(stderr, PROGNAME": packet length = %d\n", packetlen);
	}

	// we use pointers into the structure so that we don't get silly alignment problems
	eth = 	(struct ethhdr*)packetdata;
	ip =	(struct iphdr*)(eth + 1);
	icmp =	(struct icmp*)(ip + 1);

	data =	(char*)(icmp + 1);

	// get a bunch of data about the bound socket to use in building the packet
	// MAC address (wtf does `MAC' stand for anyway?)
	memset(&ifr, 0, sizeof(ifr));

	strncpy(ifr.ifr_name, if_name, IFNAMSIZ);
	ifr.ifr_hwaddr.sa_family = AF_UNSPEC;
	if (ioctl(sock, SIOCGIFADDR, &ifr) < 0)
	{
		perror(PROGNAME": ioctl");
		return 3;
	}
	memcpy(source_macaddr, ifr.ifr_hwaddr.sa_data, MACADDR_LEN);

	if (verbose)	
	{
		fprintf(stderr, PROGNAME": %s: MAC address %s, ",
				ifr.ifr_name,
				hwaddr_to_string(ifr.ifr_hwaddr.sa_data));
	}

	// local address
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, if_name, IFNAMSIZ);
	ifr.ifr_hwaddr.sa_family = AF_INET;
	if (ioctl(sock, SIOCGIFADDR, &ifr) < 0)
	{
		perror(PROGNAME": ioctl");
		return 3;
	}

	sourceaddr = ((struct sockaddr_in*)&(ifr.ifr_addr))->sin_addr.s_addr;

	if (verbose)
	{
		fprintf(stderr, "IP address %s\n",
				ipaddr_to_string(sourceaddr));
	}

	// broadcast address, if required
	if (broadcast)
	{
		memset(&ifr, 0, sizeof(ifr));
		strncpy(ifr.ifr_name, if_name, IFNAMSIZ);
		ifr.ifr_hwaddr.sa_family = AF_INET;
		if (ioctl(sock, SIOCGIFBRDADDR, &ifr) < 0)
		{
			perror(PROGNAME": ioctl");
			return 3;
		}

		destaddr = ((struct sockaddr_in*)&(ifr.ifr_broadaddr))->sin_addr.s_addr;
	}

	if (verbose)	
	{
		fprintf(stderr, PROGNAME": sending to %saddress %s\n",
			broadcast ? "broadcast " : destaddr ? "" : "`local broadcast' ",
				ipaddr_to_string(destaddr));
	}

	// set up the ethernet header
	// copy mac addresses into ethernet header

	// the source address is real... why not?
	memcpy((char *)&eth->h_source, source_macaddr, MACADDR_LEN);

	// we send to "42:42:42:42:42:42"; the probability that this MAC address
	// will be in use on a complete (256 host) randomly-populated subnet is
	// approximately .00000000000090949376, which seems a reasonable risk
	memset(&(eth->h_dest), 0x42, MACADDR_LEN);
	eth->h_proto =	htons(ETH_P_IP);

	// setup the IP header
	ip->version =		IPVERSION;	// IPv4
	ip->ihl = 		sizeof(struct iphdr) >> 2;
	ip->tos =		0;
	ip->tot_len =		htons(packetlen - sizeof(struct ethhdr));
	ip->id =		htons(getpid() & 0xffff);
	ip->frag_off =		0;
	ip->ttl =		0x40;
	ip->protocol =		IPPROTO_ICMP;
	ip->check =		0;

	// put the right IP addresses in
	ip->saddr =		sourceaddr;
	ip->daddr =		destaddr;

	ip->check =		ip_checksum((unsigned char*)ip, sizeof(struct iphdr));

	// now make up the ICMP header
	icmp->icmp_type =	ICMP_ECHO;
	icmp->icmp_code =	0;
	icmp->icmp_cksum =	0;
	icmp->icmp_seq =	htons(42);
	icmp->icmp_id =		getpid();

	memset(data, 0, DATA_LEN);
	strcpy(data, string);

	icmp->icmp_cksum = ip_checksum((unsigned char*)icmp, DATA_LEN + sizeof(struct icmp));

	// finish setting up ICMP_ECHO packet --->

	// set up the ICMP socket with a huge receive buffer; we could get many
	// responses from a broadcast ping.
	if (setsockopt(icmpsock, SOL_SOCKET, SO_RCVBUF, (char*)&buffersize, sizeof(buffersize)) < 0)
	{
		perror(PROGNAME": setsockopt");
		return 2;
	}

	// start things going
	signal(SIGALRM, pinger);

	pinger(SIGALRM);

	// now we just retrieve packets from the icmp socket, and filter out
	// those which are not ours
	memset(&sa_from, 0, sizeof(sa_from));
	icmpreply = (struct icmp*)(recvdata + sizeof(struct iphdr));

	for (;;)
	{
		n = sizeof(sa_from);

		i = recvfrom(icmpsock, recvdata, RECV_BUFFER_LEN, 0, (struct sockaddr*)&sa_from, &n);

		if (i < 0)
		{
			if (errno != EINTR)
				perror(PROGNAME": recvfrom");
		}
		else
		{
			// check whether this is indeed our packet, and is an
			// echo response
			if (icmpreply->icmp_type == ICMP_ECHOREPLY && icmpreply->icmp_id == icmp->icmp_id)
				printf("<<< got reply from %s\n", ipaddr_to_string(sa_from.sin_addr.s_addr));
		}
	}

	return 0;
}
